"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var Type_1 = require("./Type");
var Support_1 = require("./Support");
var MIN_LENGTH_FOR_ENUM = 10;
function shouldBeEnum(t) {
    var enumCases = t.enumCases;
    if (enumCases !== undefined) {
        Support_1.assert(enumCases.size > 0, "How did we end up with zero enum cases?");
        var numValues = enumCases.map(function (n) { return n; }).reduce(function (a, b) { return a + b; });
        if (numValues >= MIN_LENGTH_FOR_ENUM && enumCases.size < Math.sqrt(numValues)) {
            return t.enumCases;
        }
    }
    return undefined;
}
function replaceString(setOfOneString, builder) {
    Support_1.assert(setOfOneString.size === 1);
    var t = Support_1.defined(setOfOneString.first());
    var maybeEnumCases = shouldBeEnum(t);
    if (maybeEnumCases !== undefined) {
        return builder.getEnumType(t.getNames(), maybeEnumCases.keySeq().toOrderedSet());
    }
    return builder.getStringType(t.getNames(), undefined);
}
exports.replaceString = replaceString;
function inferEnums(graph, stringTypeMapping) {
    var allStrings = graph
        .allTypesUnordered()
        .filter(function (t) { return t instanceof Type_1.StringType && shouldBeEnum(t) !== undefined; })
        .map(function (t) { return [t]; })
        .toArray();
    return graph.rewrite(stringTypeMapping, allStrings, replaceString);
}
exports.inferEnums = inferEnums;
