"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Support_1 = require("./Support");
var Type = /** @class */ (function () {
    function Type(typeRef, kind) {
        var _this = this;
        this.typeRef = typeRef;
        this.kind = kind;
        this.getNames = function () {
            return Support_1.defined(_this.typeRef.deref()[1]);
        };
        this.getCombinedName = function () {
            return _this.getNames().combinedName;
        };
        this.getProposedNames = function () {
            return _this.getNames().proposedNames;
        };
    }
    Object.defineProperty(Type.prototype, "isStringType", {
        get: function () {
            return false;
        },
        enumerable: true,
        configurable: true
    });
    Type.prototype.directlyReachableTypes = function (setForType) {
        var set = setForType(this);
        if (set)
            return set;
        return orderedSetUnion(this.children.map(function (t) { return t.directlyReachableTypes(setForType); }));
    };
    Object.defineProperty(Type.prototype, "hasNames", {
        get: function () {
            return this.typeRef.deref()[1] !== undefined;
        },
        enumerable: true,
        configurable: true
    });
    Type.prototype.equals = function (other) {
        if (!Object.prototype.hasOwnProperty.call(other, "typeRef")) {
            return false;
        }
        return this.typeRef.equals(other.typeRef);
    };
    Type.prototype.hashCode = function () {
        return this.typeRef.hashCode();
    };
    return Type;
}());
exports.Type = Type;
var PrimitiveType = /** @class */ (function (_super) {
    __extends(PrimitiveType, _super);
    function PrimitiveType(typeRef, kind) {
        return _super.call(this, typeRef, kind) || this;
    }
    Object.defineProperty(PrimitiveType.prototype, "children", {
        get: function () {
            return immutable_1.OrderedSet();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(PrimitiveType.prototype, "isNullable", {
        get: function () {
            return this.kind === "null";
        },
        enumerable: true,
        configurable: true
    });
    PrimitiveType.prototype.isPrimitive = function () {
        return true;
    };
    Object.defineProperty(PrimitiveType.prototype, "isStringType", {
        get: function () {
            var kind = this.kind;
            return kind === "string" || kind === "date" || kind === "time" || kind === "date-time";
        },
        enumerable: true,
        configurable: true
    });
    PrimitiveType.prototype.map = function (builder, _) {
        return builder.getPrimitiveType(this.kind);
    };
    return PrimitiveType;
}(Type));
exports.PrimitiveType = PrimitiveType;
function isNull(t) {
    return t.kind === "null";
}
var StringType = /** @class */ (function (_super) {
    __extends(StringType, _super);
    function StringType(typeRef, enumCases) {
        var _this = _super.call(this, typeRef, "string") || this;
        _this.enumCases = enumCases;
        return _this;
    }
    StringType.prototype.map = function (builder, _) {
        return builder.getStringType(this.enumCases);
    };
    return StringType;
}(PrimitiveType));
exports.StringType = StringType;
var ArrayType = /** @class */ (function (_super) {
    __extends(ArrayType, _super);
    function ArrayType(typeRef, _itemsRef) {
        var _this = _super.call(this, typeRef, "array") || this;
        _this._itemsRef = _itemsRef;
        return _this;
    }
    ArrayType.prototype.setItems = function (itemsRef) {
        if (this._itemsRef !== undefined) {
            return Support_1.panic("Can only set array items once");
        }
        this._itemsRef = itemsRef;
    };
    ArrayType.prototype.getItemsRef = function () {
        if (this._itemsRef === undefined) {
            return Support_1.panic("Array items accessed before they were set");
        }
        return this._itemsRef;
    };
    Object.defineProperty(ArrayType.prototype, "items", {
        get: function () {
            return this.getItemsRef().deref()[0];
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ArrayType.prototype, "children", {
        get: function () {
            return immutable_1.OrderedSet([this.items]);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ArrayType.prototype, "isNullable", {
        get: function () {
            return false;
        },
        enumerable: true,
        configurable: true
    });
    ArrayType.prototype.isPrimitive = function () {
        return false;
    };
    ArrayType.prototype.map = function (builder, f) {
        return builder.getArrayType(f(this.getItemsRef()));
    };
    return ArrayType;
}(Type));
exports.ArrayType = ArrayType;
var MapType = /** @class */ (function (_super) {
    __extends(MapType, _super);
    function MapType(typeRef, _valuesRef) {
        var _this = _super.call(this, typeRef, "map") || this;
        _this._valuesRef = _valuesRef;
        return _this;
    }
    MapType.prototype.setValues = function (valuesRef) {
        if (this._valuesRef !== undefined) {
            return Support_1.panic("Can only set map values once");
        }
        this._valuesRef = valuesRef;
    };
    MapType.prototype.getValuesRef = function () {
        if (this._valuesRef === undefined) {
            return Support_1.panic("Map values accessed before they were set");
        }
        return this._valuesRef;
    };
    Object.defineProperty(MapType.prototype, "values", {
        get: function () {
            return this.getValuesRef().deref()[0];
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(MapType.prototype, "children", {
        get: function () {
            return immutable_1.OrderedSet([this.values]);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(MapType.prototype, "isNullable", {
        get: function () {
            return false;
        },
        enumerable: true,
        configurable: true
    });
    MapType.prototype.isPrimitive = function () {
        return false;
    };
    MapType.prototype.map = function (builder, f) {
        return builder.getMapType(f(this.getValuesRef()));
    };
    return MapType;
}(Type));
exports.MapType = MapType;
var ClassType = /** @class */ (function (_super) {
    __extends(ClassType, _super);
    function ClassType(typeRef, isFixed, _propertyRefs) {
        var _this = _super.call(this, typeRef, "class") || this;
        _this.isFixed = isFixed;
        _this._propertyRefs = _propertyRefs;
        return _this;
    }
    ClassType.prototype.setProperties = function (propertyRefs) {
        if (this._propertyRefs !== undefined) {
            return Support_1.panic("Can only set class properties once");
        }
        this._propertyRefs = propertyRefs;
    };
    ClassType.prototype.getPropertyRefs = function () {
        if (this._propertyRefs === undefined) {
            return Support_1.panic("Class properties accessed before they were set");
        }
        return this._propertyRefs;
    };
    Object.defineProperty(ClassType.prototype, "properties", {
        get: function () {
            return this.getPropertyRefs().map(function (tref) { return tref.deref()[0]; });
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ClassType.prototype, "sortedProperties", {
        get: function () {
            var properties = this.properties;
            var sortedKeys = properties.keySeq().sort();
            var props = sortedKeys.map(function (k) { return [k, Support_1.defined(properties.get(k))]; });
            return immutable_1.OrderedMap(props);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ClassType.prototype, "children", {
        get: function () {
            return this.sortedProperties.toOrderedSet();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ClassType.prototype, "isNullable", {
        get: function () {
            return false;
        },
        enumerable: true,
        configurable: true
    });
    ClassType.prototype.isPrimitive = function () {
        return false;
    };
    ClassType.prototype.map = function (builder, f) {
        var properties = this.getPropertyRefs().map(f);
        if (this.isFixed) {
            return builder.getUniqueClassType(properties);
        }
        else {
            return builder.getClassType(properties);
        }
    };
    return ClassType;
}(Type));
exports.ClassType = ClassType;
var EnumType = /** @class */ (function (_super) {
    __extends(EnumType, _super);
    function EnumType(typeRef, cases) {
        var _this = _super.call(this, typeRef, "enum") || this;
        _this.cases = cases;
        return _this;
    }
    Object.defineProperty(EnumType.prototype, "children", {
        get: function () {
            return immutable_1.OrderedSet();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(EnumType.prototype, "isNullable", {
        get: function () {
            return false;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(EnumType.prototype, "isStringType", {
        get: function () {
            return true;
        },
        enumerable: true,
        configurable: true
    });
    EnumType.prototype.isPrimitive = function () {
        return false;
    };
    EnumType.prototype.map = function (builder, _) {
        return builder.getEnumType(this.cases);
    };
    return EnumType;
}(Type));
exports.EnumType = EnumType;
var UnionType = /** @class */ (function (_super) {
    __extends(UnionType, _super);
    function UnionType(typeRef, _memberRefs) {
        var _this = _super.call(this, typeRef, "union") || this;
        _this._memberRefs = _memberRefs;
        _this.findMember = function (kind) {
            return _this.members.find(function (t) { return t.kind === kind; });
        };
        if (_memberRefs !== undefined) {
            Support_1.assert(_memberRefs.size > 1, "Union has zero members");
        }
        return _this;
    }
    UnionType.prototype.setMembers = function (memberRefs) {
        if (this._memberRefs !== undefined) {
            return Support_1.panic("Can only set map members once");
        }
        Support_1.assert(memberRefs.size > 1, "Union has zero members");
        this._memberRefs = memberRefs;
    };
    UnionType.prototype.getMemberRefs = function () {
        if (this._memberRefs === undefined) {
            return Support_1.panic("Map members accessed before they were set");
        }
        return this._memberRefs;
    };
    Object.defineProperty(UnionType.prototype, "members", {
        get: function () {
            return this.getMemberRefs().map(function (tref) { return tref.deref()[0]; });
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(UnionType.prototype, "stringTypeMembers", {
        get: function () {
            return this.members.filter(function (t) { return t.isStringType; });
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(UnionType.prototype, "children", {
        get: function () {
            return this.sortedMembers;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(UnionType.prototype, "isNullable", {
        get: function () {
            return this.findMember("null") !== undefined;
        },
        enumerable: true,
        configurable: true
    });
    UnionType.prototype.isPrimitive = function () {
        return false;
    };
    UnionType.prototype.map = function (builder, f) {
        var members = this.getMemberRefs().map(f);
        return builder.getUnionType(members);
    };
    Object.defineProperty(UnionType.prototype, "sortedMembers", {
        get: function () {
            // FIXME: We're assuming no two members of the same kind.
            return this.members.sortBy(function (t) { return t.kind; });
        },
        enumerable: true,
        configurable: true
    });
    return UnionType;
}(Type));
exports.UnionType = UnionType;
function removeNullFromUnion(t) {
    var nullType = t.findMember("null");
    if (!nullType) {
        return [null, t.members];
    }
    return [nullType, t.members.filterNot(isNull).toOrderedSet()];
}
exports.removeNullFromUnion = removeNullFromUnion;
function nullableFromUnion(t) {
    var _a = removeNullFromUnion(t), hasNull = _a[0], nonNulls = _a[1];
    if (!hasNull)
        return null;
    if (nonNulls.size !== 1)
        return null;
    return Support_1.defined(nonNulls.first());
}
exports.nullableFromUnion = nullableFromUnion;
function nonNullTypeCases(t) {
    if (t.kind === "null") {
        return immutable_1.OrderedSet();
    }
    if (!(t instanceof UnionType)) {
        return immutable_1.OrderedSet([t]);
    }
    var nonNulls = removeNullFromUnion(t)[1];
    return immutable_1.OrderedSet(nonNulls);
}
exports.nonNullTypeCases = nonNullTypeCases;
// FIXME: The outer OrderedSet should be some Collection, but I can't figure out
// which one.  Collection.Indexed doesn't work with OrderedSet, which is unfortunate.
function orderedSetUnion(sets) {
    var setArray = sets.toArray();
    if (setArray.length === 0)
        return immutable_1.OrderedSet();
    if (setArray.length === 1)
        return setArray[0];
    return (_a = setArray[0]).union.apply(_a, setArray.slice(1));
    var _a;
}
// FIXME: Give this an appropriate name, considering that we don't distinguish
// between named and non-named types anymore.
function isNamedType(t) {
    return t instanceof ClassType || t instanceof EnumType || t instanceof UnionType;
}
exports.isNamedType = isNamedType;
function separateNamedTypes(types) {
    var classes = types.filter(function (t) { return t instanceof ClassType; }).toOrderedSet();
    var enums = types.filter(function (t) { return t instanceof EnumType; }).toOrderedSet();
    var unions = types.filter(function (t) { return t instanceof UnionType; }).toOrderedSet();
    return { classes: classes, enums: enums, unions: unions };
}
exports.separateNamedTypes = separateNamedTypes;
function matchTypeExhaustive(t, anyType, nullType, boolType, integerType, doubleType, stringType, arrayType, classType, mapType, enumType, unionType, dateType, timeType, dateTimeType) {
    if (t.isPrimitive()) {
        var f = {
            any: anyType,
            null: nullType,
            bool: boolType,
            integer: integerType,
            double: doubleType,
            string: stringType,
            date: dateType,
            time: timeType,
            "date-time": dateTimeType
        }[t.kind];
        if (f)
            return f(t);
        return Support_1.panic("Unsupported PrimitiveType: " + t.kind);
    }
    else if (t instanceof ArrayType)
        return arrayType(t);
    else if (t instanceof ClassType)
        return classType(t);
    else if (t instanceof MapType)
        return mapType(t);
    else if (t instanceof EnumType)
        return enumType(t);
    else if (t instanceof UnionType)
        return unionType(t);
    return Support_1.panic("Unknown Type");
}
exports.matchTypeExhaustive = matchTypeExhaustive;
function matchType(t, anyType, nullType, boolType, integerType, doubleType, stringType, arrayType, classType, mapType, enumType, unionType, stringTypeMatchers) {
    if (stringTypeMatchers === undefined) {
        stringTypeMatchers = {};
    }
    var typeNotSupported = function (_) {
        return Support_1.panic("Unsupported PrimitiveType");
    };
    return matchTypeExhaustive(t, anyType, nullType, boolType, integerType, doubleType, stringType, arrayType, classType, mapType, enumType, unionType, stringTypeMatchers.dateType || typeNotSupported, stringTypeMatchers.timeType || typeNotSupported, stringTypeMatchers.dateTimeType || typeNotSupported);
}
exports.matchType = matchType;
function matchCompoundType(t, arrayType, classType, mapType, unionType) {
    function ignore(_) {
        return;
    }
    return matchType(t, ignore, ignore, ignore, ignore, ignore, ignore, arrayType, classType, mapType, ignore, unionType, { dateType: ignore, timeType: ignore, dateTimeType: ignore });
}
exports.matchCompoundType = matchCompoundType;
