"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Type_1 = require("./Type");
var TypeGraph_1 = require("./TypeGraph");
var Support_1 = require("./Support");
var TypeRef = /** @class */ (function () {
    function TypeRef(graph, index, _allocate) {
        var _this = this;
        this.graph = graph;
        this._allocate = _allocate;
        this.callWhenResolved = function (callback) {
            if (_this._maybeIndexOrRef === undefined) {
                if (_this._callbacks === undefined) {
                    _this._callbacks = [];
                }
                _this._callbacks.push(callback);
            }
            else if (typeof _this._maybeIndexOrRef === "number") {
                callback(_this._maybeIndexOrRef);
            }
            else {
                _this._maybeIndexOrRef.callWhenResolved(callback);
            }
        };
        this.resolve = function (tref) {
            if (_this._maybeIndexOrRef !== undefined) {
                Support_1.assert(_this.maybeIndex === tref.maybeIndex, "Trying to resolve an allocated type reference with an incompatible one");
            }
            else {
                Support_1.assert(tref.follow() !== _this, "Tried to create a TypeRef cycle");
            }
            _this._maybeIndexOrRef = tref;
            _this._allocate = undefined;
            if (_this._callbacks !== undefined) {
                for (var _i = 0, _a = _this._callbacks; _i < _a.length; _i++) {
                    var cb = _a[_i];
                    tref.callWhenResolved(cb);
                }
                _this._callbacks = undefined;
            }
        };
        this.deref = function () {
            return _this.graph.atIndex(_this.index);
        };
        this.equals = function (other) {
            if (!(other instanceof TypeRef)) {
                return false;
            }
            Support_1.assert(_this.graph === other.graph, "Comparing type refs of different graphs");
            return _this.follow() === other.follow();
        };
        this.hashCode = function () {
            return _this.index | 0;
        };
        this._maybeIndexOrRef = index;
    }
    TypeRef.prototype.follow = function () {
        if (this._maybeIndexOrRef instanceof TypeRef) {
            return this._maybeIndexOrRef.follow();
        }
        return this;
    };
    Object.defineProperty(TypeRef.prototype, "maybeIndex", {
        get: function () {
            var tref = this.follow();
            if (typeof tref._maybeIndexOrRef === "number") {
                return tref._maybeIndexOrRef;
            }
            return undefined;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TypeRef.prototype, "index", {
        get: function () {
            var maybeIndex = this.maybeIndex;
            if (maybeIndex === undefined) {
                var tref = this.follow();
                if (tref._allocate !== undefined) {
                    var allocated = tref._allocate();
                    tref._maybeIndexOrRef = allocated;
                    tref._allocate = undefined;
                    return allocated.index;
                }
                return Support_1.panic("Trying to dereference unresolved type reference");
            }
            return maybeIndex;
        },
        enumerable: true,
        configurable: true
    });
    return TypeRef;
}());
exports.TypeRef = TypeRef;
var TypeBuilder = /** @class */ (function () {
    function TypeBuilder(_stringTypeMapping) {
        var _this = this;
        this._stringTypeMapping = _stringTypeMapping;
        this.typeGraph = new TypeGraph_1.TypeGraph(this);
        this.topLevels = immutable_1.Map();
        this.types = immutable_1.List();
        this.typeNames = immutable_1.List();
        this.addTopLevel = function (name, tref) {
            // assert(t.typeGraph === this.typeGraph, "Adding top-level to wrong type graph");
            Support_1.assert(!_this.topLevels.has(name), "Trying to add top-level with existing name");
            Support_1.assert(_this.types.get(tref.index) !== undefined, "Trying to add a top-level type that doesn't exist (yet?)");
            _this.topLevels = _this.topLevels.set(name, tref);
        };
        this.reserveTypeRef = function () {
            var index = _this.types.size;
            _this.types = _this.types.push(undefined);
            _this.typeNames = _this.typeNames.push(undefined);
            return new TypeRef(_this.typeGraph, index, undefined);
        };
        this.commitType = function (tref, t, names) {
            Support_1.assert(_this.types.get(tref.index) === undefined, "A type index was committed twice");
            _this.types = _this.types.set(tref.index, t);
            _this.typeNames = _this.typeNames.set(tref.index, names);
        };
        this.atIndex = function (index) {
            var maybeType = _this.types.get(index);
            if (maybeType === undefined) {
                return Support_1.panic("Trying to deref an undefined type in a type builder");
            }
            var maybeNames = _this.typeNames.get(index);
            return [maybeType, maybeNames];
        };
        this.addNames = function (tref, names) {
            tref.callWhenResolved(function (index) {
                var tn = _this.typeNames.get(index);
                if (tn === undefined) {
                    _this.typeNames = _this.typeNames.set(index, names);
                }
                else {
                    tn.add(names);
                }
            });
        };
        this.makeNullable = function (tref, typeNames) {
            var t = Support_1.defined(_this.types.get(tref.index));
            if (t.kind === "null") {
                return tref;
            }
            var nullType = _this.getPrimitiveType("null");
            if (!(t instanceof Type_1.UnionType)) {
                return _this.getUnionType(typeNames, immutable_1.OrderedSet([tref, nullType]));
            }
            var _a = Type_1.removeNullFromUnion(t), maybeNull = _a[0], nonNulls = _a[1];
            if (maybeNull)
                return tref;
            return _this.getUnionType(typeNames, nonNulls.map(function (nn) { return nn.typeRef; }).add(nullType));
        };
        // FIXME: make mutable?
        this._primitiveTypes = immutable_1.Map();
        this._noEnumStringType = undefined;
        this._mapTypes = immutable_1.Map();
        this._arrayTypes = immutable_1.Map();
        this._enumTypes = immutable_1.Map();
        this._classTypes = immutable_1.Map();
        this._unionTypes = immutable_1.Map();
        // FIXME: Maybe just distinguish between this and `getClassType`
        // via a flag?  That would make `ClassType.map` simpler.
        this.getUniqueClassType = function (names, properties, forwardingRef) {
            return _this.addType(forwardingRef, function (tref) { return new Type_1.ClassType(tref, true, properties); }, names);
        };
    }
    TypeBuilder.prototype.addType = function (forwardingRef, creator, names) {
        if (names !== undefined) {
            // We need to copy the names here because they're modified
            // in `gatherNames`, and the caller doesn't guarantee that
            // this one is unique for this type.
            names = names.copy();
        }
        var tref = forwardingRef !== undefined && forwardingRef.maybeIndex !== undefined
            ? forwardingRef
            : this.reserveTypeRef();
        if (names !== undefined) {
            this.addNames(tref, names);
        }
        var t = creator(tref);
        this.commitType(tref, t, names);
        if (forwardingRef !== undefined && tref !== forwardingRef) {
            forwardingRef.resolve(tref);
        }
        return tref;
    };
    TypeBuilder.prototype.finish = function () {
        this.typeGraph.freeze(this.topLevels, this.types.map(Support_1.defined), this.typeNames);
        return this.typeGraph;
    };
    TypeBuilder.prototype.getPrimitiveType = function (kind, forwardingRef) {
        Support_1.assert(kind !== "string", "Use getStringType to create strings");
        if (kind === "date")
            kind = this._stringTypeMapping.date;
        if (kind === "time")
            kind = this._stringTypeMapping.time;
        if (kind === "date-time")
            kind = this._stringTypeMapping.dateTime;
        var tref = this._primitiveTypes.get(kind);
        if (tref === undefined) {
            tref = this.addType(forwardingRef, function (tr) { return new Type_1.PrimitiveType(tr, kind); }, undefined);
            this._primitiveTypes = this._primitiveTypes.set(kind, tref);
        }
        return tref;
    };
    TypeBuilder.prototype.getStringType = function (names, cases, forwardingRef) {
        if (names === undefined && cases === undefined) {
            if (this._noEnumStringType === undefined) {
                this._noEnumStringType = this.addType(forwardingRef, function (tr) { return new Type_1.StringType(tr, undefined); }, undefined);
            }
            return this._noEnumStringType;
        }
        return this.addType(forwardingRef, function (tr) { return new Type_1.StringType(tr, cases); }, names);
    };
    TypeBuilder.prototype.getEnumType = function (names, cases, forwardingRef) {
        var unorderedCases = cases.toSet();
        var tref = this._enumTypes.get(unorderedCases);
        if (tref === undefined) {
            tref = this.addType(forwardingRef, function (tr) { return new Type_1.EnumType(tr, cases); }, names);
            this._enumTypes = this._enumTypes.set(unorderedCases, tref);
        }
        else {
            this.addNames(tref, names);
        }
        return tref;
    };
    TypeBuilder.prototype.getMapType = function (values, forwardingRef) {
        var tref = this._mapTypes.get(values);
        if (tref === undefined) {
            tref = this.addType(forwardingRef, function (tr) { return new Type_1.MapType(tr, values); }, undefined);
            this._mapTypes = this._mapTypes.set(values, tref);
        }
        return tref;
    };
    TypeBuilder.prototype.getArrayType = function (items, forwardingRef) {
        var tref = this._arrayTypes.get(items);
        if (tref === undefined) {
            tref = this.addType(forwardingRef, function (tr) { return new Type_1.ArrayType(tr, items); }, undefined);
            this._arrayTypes = this._arrayTypes.set(items, tref);
        }
        return tref;
    };
    TypeBuilder.prototype.getClassType = function (names, properties, forwardingRef) {
        var tref = this._classTypes.get(properties.toMap());
        if (forwardingRef !== undefined || tref === undefined) {
            tref = this.addType(forwardingRef, function (tr) { return new Type_1.ClassType(tr, false, properties); }, names);
            this._classTypes = this._classTypes.set(properties.toMap(), tref);
        }
        else {
            this.addNames(tref, names);
        }
        return tref;
    };
    TypeBuilder.prototype.getUnionType = function (names, members, forwardingRef) {
        var unorderedMembers = members.toSet();
        var tref = this._unionTypes.get(unorderedMembers);
        if (tref === undefined) {
            tref = this.addType(forwardingRef, function (tr) { return new Type_1.UnionType(tr, members); }, names);
            this._unionTypes = this._unionTypes.set(unorderedMembers, tref);
        }
        else {
            this.addNames(tref, names);
        }
        return tref;
    };
    return TypeBuilder;
}());
exports.TypeBuilder = TypeBuilder;
var TypeGraphBuilder = /** @class */ (function (_super) {
    __extends(TypeGraphBuilder, _super);
    function TypeGraphBuilder() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.getUniqueUnionType = function (names, members) {
            return _this.addType(undefined, function (tref) { return new Type_1.UnionType(tref, members); }, names);
        };
        _this.lookupType = function (typeRef) {
            var maybeIndex = typeRef.maybeIndex;
            if (maybeIndex === undefined) {
                return undefined;
            }
            return _this.types.get(maybeIndex);
        };
        return _this;
    }
    TypeGraphBuilder.prototype.typeForEntry = function (entry) {
        return entry;
    };
    TypeGraphBuilder.prototype.getLazyMapType = function (valuesCreator) {
        return this.addType(undefined, function (tref) { return new Type_1.MapType(tref, valuesCreator()); }, undefined);
    };
    return TypeGraphBuilder;
}(TypeBuilder));
exports.TypeGraphBuilder = TypeGraphBuilder;
var TypeReconstituter = /** @class */ (function () {
    function TypeReconstituter(_typeBuilder, _typeNames, _forwardingRef) {
        var _this = this;
        this._typeBuilder = _typeBuilder;
        this._typeNames = _typeNames;
        this._forwardingRef = _forwardingRef;
        this._wasUsed = false;
        this.useBuilder = function () {
            Support_1.assert(!_this._wasUsed, "TypeReconstituter used more than once");
            _this._wasUsed = true;
            return _this._typeBuilder;
        };
        this.getPrimitiveType = function (kind) {
            return _this.useBuilder().getPrimitiveType(kind, _this._forwardingRef);
        };
        this.getStringType = function (enumCases) {
            return _this.useBuilder().getStringType(Support_1.defined(_this._typeNames), enumCases, _this._forwardingRef);
        };
        this.getEnumType = function (cases) {
            return _this.useBuilder().getEnumType(Support_1.defined(_this._typeNames), cases, _this._forwardingRef);
        };
        this.getMapType = function (values) {
            return _this.useBuilder().getMapType(values, _this._forwardingRef);
        };
        this.getArrayType = function (items) {
            return _this.useBuilder().getArrayType(items, _this._forwardingRef);
        };
        this.getClassType = function (properties) {
            return _this.useBuilder().getClassType(Support_1.defined(_this._typeNames), properties, _this._forwardingRef);
        };
        this.getUniqueClassType = function (properties) {
            return _this.useBuilder().getUniqueClassType(Support_1.defined(_this._typeNames), properties, _this._forwardingRef);
        };
        this.getUnionType = function (members) {
            return _this.useBuilder().getUnionType(Support_1.defined(_this._typeNames), members, _this._forwardingRef);
        };
    }
    return TypeReconstituter;
}());
exports.TypeReconstituter = TypeReconstituter;
var GraphRewriteBuilder = /** @class */ (function (_super) {
    __extends(GraphRewriteBuilder, _super);
    function GraphRewriteBuilder(_originalGraph, stringTypeMapping, setsToReplace, _replacer) {
        var _this = _super.call(this, stringTypeMapping) || this;
        _this._originalGraph = _originalGraph;
        _this._replacer = _replacer;
        _this._reconstitutedTypes = immutable_1.Map();
        _this.getReconstitutedType = function (originalRef) {
            var index = originalRef.index;
            var maybeTypeRef = _this._reconstitutedTypes.get(index);
            if (maybeTypeRef !== undefined) {
                return maybeTypeRef;
            }
            var maybeSet = _this._setsToReplaceByMember.get(index);
            if (maybeSet !== undefined) {
                return _this.replaceSet(maybeSet);
            }
            return _this.withForwardingRef(function (forwardingRef) {
                _this._reconstitutedTypes = _this._reconstitutedTypes.set(index, forwardingRef);
                var _a = originalRef.deref(), originalType = _a[0], originalNames = _a[1];
                return originalType.map(new TypeReconstituter(_this, originalNames, forwardingRef), _this.getReconstitutedType);
            });
        };
        _this.reconstituteType = function (t) {
            Support_1.assert(t.typeRef.graph === _this._originalGraph, "Trying to reconstitute a type from the wrong graph");
            return _this.getReconstitutedType(t.typeRef);
        };
        _this._setsToReplaceByMember = immutable_1.Map();
        var _loop_1 = function (types) {
            var set = immutable_1.Set(types);
            set.forEach(function (t) {
                var index = t.typeRef.index;
                Support_1.assert(!_this._setsToReplaceByMember.has(index), "A type is member of more than one set to be replaced");
                _this._setsToReplaceByMember = _this._setsToReplaceByMember.set(index, set);
            });
        };
        for (var _i = 0, setsToReplace_1 = setsToReplace; _i < setsToReplace_1.length; _i++) {
            var types = setsToReplace_1[_i];
            _loop_1(types);
        }
        return _this;
    }
    GraphRewriteBuilder.prototype.followIndex = function (index) {
        var entry = this.types.get(index);
        if (typeof entry === "number") {
            return this.followIndex(entry);
        }
        return index;
    };
    GraphRewriteBuilder.prototype.typeForEntry = function (entry) {
        if (typeof entry === "number") {
            entry = this.types.get(this.followIndex(entry));
            if (typeof entry === "number") {
                return Support_1.panic("followIndex led us to a forwarding entry");
            }
        }
        return entry;
    };
    GraphRewriteBuilder.prototype.withForwardingRef = function (typeCreator) {
        var forwardingRef = new TypeRef(this.typeGraph, undefined, this.reserveTypeRef);
        var actualRef = typeCreator(forwardingRef);
        forwardingRef.resolve(actualRef);
        return actualRef;
    };
    GraphRewriteBuilder.prototype.replaceSet = function (typesToReplace) {
        var _this = this;
        return this.withForwardingRef(function (forwardingRef) {
            typesToReplace.forEach(function (t) {
                var originalRef = t.typeRef;
                _this._reconstitutedTypes = _this._reconstitutedTypes.set(originalRef.index, forwardingRef);
                _this._setsToReplaceByMember = _this._setsToReplaceByMember.remove(originalRef.index);
            });
            return _this._replacer(typesToReplace, _this);
        });
    };
    GraphRewriteBuilder.prototype.finish = function () {
        var _this = this;
        this._originalGraph.topLevels.forEach(function (t, name) {
            _this.addTopLevel(name, _this.getReconstitutedType(t.typeRef));
        });
        return _super.prototype.finish.call(this);
    };
    return GraphRewriteBuilder;
}(TypeBuilder));
exports.GraphRewriteBuilder = GraphRewriteBuilder;
var UnionBuilder = /** @class */ (function () {
    function UnionBuilder(typeBuilder, typeNames) {
        var _this = this;
        this.typeBuilder = typeBuilder;
        this.typeNames = typeNames;
        this._haveAny = false;
        this._haveNull = false;
        this._haveBool = false;
        this._haveInteger = false;
        this._haveDouble = false;
        this._stringTypes = immutable_1.OrderedSet();
        this._arrays = [];
        this._maps = [];
        this._classes = [];
        this._enumCaseMap = {};
        this._enumCases = [];
        this.addAny = function () {
            _this._haveAny = true;
        };
        this.addNull = function () {
            _this._haveNull = true;
        };
        this.addBool = function () {
            _this._haveBool = true;
        };
        this.addInteger = function () {
            _this._haveInteger = true;
        };
        this.addDouble = function () {
            _this._haveDouble = true;
        };
        this.addStringType = function (kind) {
            if (_this._stringTypes.has(kind))
                return;
            // string overrides all other string types, as well as enum
            if (kind === "string") {
                _this._stringTypes = immutable_1.OrderedSet([kind]);
                _this._enumCaseMap = {};
                _this._enumCases = [];
            }
            else {
                if (_this.haveString)
                    return;
                _this._stringTypes = _this._stringTypes.add(kind);
            }
        };
        this.addArray = function (t) {
            _this._arrays.push(t);
        };
        this.addClass = function (t) {
            _this._classes.push(t);
        };
        this.addMap = function (t) {
            _this._maps.push(t);
        };
        this.addEnumCase = function (s) {
            if (_this.haveString) {
                return;
            }
            if (!Object.prototype.hasOwnProperty.call(_this._enumCaseMap, s)) {
                _this._enumCaseMap[s] = 0;
                _this._enumCases.push(s);
            }
            _this._enumCaseMap[s] += 1;
        };
        this.buildUnion = function (unique) {
            var types = [];
            if (_this._haveAny) {
                return _this.typeBuilder.getPrimitiveType("any");
            }
            if (_this._haveNull) {
                types.push(_this.typeBuilder.getPrimitiveType("null"));
            }
            if (_this._haveBool) {
                types.push(_this.typeBuilder.getPrimitiveType("bool"));
            }
            if (_this._haveDouble) {
                types.push(_this.typeBuilder.getPrimitiveType("double"));
            }
            else if (_this._haveInteger) {
                types.push(_this.typeBuilder.getPrimitiveType("integer"));
            }
            _this._stringTypes.forEach(function (kind) {
                types.push(kind === "string"
                    ? _this.typeBuilder.getStringType(_this.typeNames, undefined)
                    : _this.typeBuilder.getPrimitiveType(kind));
            });
            if (_this._enumCases.length > 0) {
                types.push(_this.makeEnum(_this._enumCases, _this._enumCaseMap));
            }
            if (_this._classes.length > 0 || _this._maps.length > 0) {
                types.push(_this.makeClass(_this._classes, _this._maps));
            }
            if (_this._arrays.length > 0) {
                types.push(_this.makeArray(_this._arrays));
            }
            if (types.length === 0) {
                return _this.typeBuilder.getPrimitiveType("any");
            }
            if (types.length === 1) {
                _this.typeBuilder.addNames(types[0], _this.typeNames);
                return types[0];
            }
            var typesSet = immutable_1.OrderedSet(types);
            if (unique) {
                return _this.typeBuilder.getUniqueUnionType(_this.typeNames, typesSet);
            }
            else {
                return _this.typeBuilder.getUnionType(_this.typeNames, typesSet);
            }
        };
    }
    Object.defineProperty(UnionBuilder.prototype, "haveString", {
        get: function () {
            return this._stringTypes.has("string");
        },
        enumerable: true,
        configurable: true
    });
    return UnionBuilder;
}());
exports.UnionBuilder = UnionBuilder;
